import logging
import threading
from typing import Dict

from serial_teleinfo import Client, Value, UnknownKeyException, ParserException

logger = logging.getLogger(__name__)

MAX_CONSECUTIVE_ERRORS = 16


class ValueUpdater:
    """ Helper class updating the values in a background thread
    and providing the result through the values property.
    """

    def __init__(self, port):
        # Config
        self._port = port

        # State
        self._values = {}
        self._connected = False
        self._ignored_keys = []  # Unknown keys already logged

        # Thread stuffs
        self._thread = None
        self._stop_event = None

    def start(self):
        """ Starts the background thread
        """
        if self._thread is not None:
            raise RuntimeError("Method start() can only be called once.")

        self._stop_event = threading.Event()
        self._thread = threading.Thread(
            target=self._update_loop, name=f"{self._port} teleinfo updater", daemon=True
        )
        self._thread.start()

    def stop(self):
        """ Stops the background thread and waits for its death.
        """
        self._stop_event.set()
        self._thread.join()

    def update_value(self, value: Value):
        """ This method can be overriden to access values soon as they are read
        """
        self._values[value.key] = value

    def _update_loop(self):
        while True:
            try:
                logger.debug(f"Opening serial connection to {self._port}")
                with Client(self._port) as client:
                    self._update(client)
            except Exception as e:
                logger.error(
                    f"Error while updating values : {e}\n\nTrying to reconnect in 10 seconds..."
                )

            logger.debug(f"Disconnected from {self._port}")
            self._connected = False

            if self._stop_event.wait(10):
                return  # We were asked to stop

    def _update(self, client):
        """ Reads the values and updates the values dict continuoulsy.

        Raises:
            TeleinfoException: If more than 16 consecutive lines coulds not be read.
            serial.SerialException: If the serial connection had an error.
        """
        consecutive_read_errors = 0

        while True:
            try:
                # Exit if the thread should stop
                if self._stop_event.is_set():
                    return

                # Actually read the value
                value = client.read_value()

                # Log if we succesfully read our first value
                if not self._connected:
                    self._connected = True
                    logger.info("Connected to Teleinfo")

                # Update table
                self.update_value(value)

                # Reset counters
                consecutive_read_errors = 0
            except UnknownKeyException as e:
                if e.key not in self._ignored_keys:
                    self._ignored_keys.append(e.key)

                    logger.warn(
                        f"Unable to find a parser for {e.key}. This key will be ignored."
                    )
            except ParserException as e:
                logger.debug(str(e))
                consecutive_read_errors += 1

                if consecutive_read_errors > MAX_CONSECUTIVE_ERRORS:
                    # Will be catched in
                    raise Exception(
                        f"Encountered too many unreadable lines while reading on {self._port}."
                    )

    @property
    def values(self) -> Dict[str, Value]:
        """ Returns a copy of the current values dictionnary
        """
        return dict(self._values)

    @property
    def port(self) -> str:
        return self._port

    @property
    def connected(self) -> bool:
        return self._connected



import configparser
import argparse
import sys
import logging
from aiohttp import web

from serial_teleinfo.server.server import app_factory

# Fetch arguments
parser = argparse.ArgumentParser()
parser.add_argument("config_path")
args = parser.parse_args()

# Parse configuration
try:
    try:
        config = configparser.ConfigParser()
        config.read(args.config_path)

        device = config["teleinfo"]["device"]
        http_listen = config["http"]["listen"]
        users = config["users"]
    except KeyError as e:
        raise Exception(f"Could not find the key {e}")

    # Parse listen field
    try:
        http_host, http_port = http_listen.split(":", 2)
        http_port = int(http_port)
    except Exception as e:
        raise Exception(f"Could not parse http/listen field {http_listen}: {e}")

    # Parse loglevel
    try:
        loglevel = getattr(logging, config["teleinfo"].get("loglevel", "INFO"))
    except Exception as e:
        raise Exception(f"Could not parse the field loglevel : {e}")
except Exception as e:
    print(
        f"Error while loading configuration, please check {args.config_path}",
        file=sys.stderr,
    )
    print(str(e), file=sys.stderr)
    sys.exit(1)


# Configure logging
logging.basicConfig(level=loglevel)

# Create the client
server = app_factory(device, users)
web.run_app(server, port=http_port, host=http_host, access_log=None)



import serial

from serial_teleinfo.parser import parse_line
from serial_teleinfo.types import type_parsers, Value
from serial_teleinfo.exception import (
    ParserException,
    UnknownKeyException,
)


class Client:
    def __init__(self, port: str = "/dev/ttyUSB0"):
        self._port = port
        self._serial = None

    def __enter__(self) -> "Client":
        """ Opens the connections to the serial port.

        Raises:
            serial.SerialException: If an error occured while establishing the connection
        """
        self._serial = serial.Serial(
            self._port,
            baudrate=1200,
            bytesize=serial.SEVENBITS,
            parity=serial.PARITY_EVEN,
            stopbits=serial.STOPBITS_ONE,
            xonxoff=False,
            timeout=1,
        )

        # Read first line as it will not be complete
        try:
            self._serial.readline()
        except:
            self._serial.close()

            raise

        return self

    def __exit__(self, type, value, traceback):
        if self._serial is not None:
            self._serial.close()

    def _raw_line(self) -> bytes:
        """ Returns next line read on the serial port
        """
        return self._serial.readline()

    def read_line(self) -> (str, str):
        """ Parses the next line and returns it as a key, value pair

        Raises:
            ParserException: If the line could not be decoded
        """
        return parse_line(self._raw_line())

    def read_value(self) -> Value:
        """ Parses the next line and returns it as a Value object
        """
        key, value, end_of_frame = self.read_line()

        # Parse the data
        parser = type_parsers.get(key)
        if parser is None:
            raise UnknownKeyException(key=key)

        return parser(key, value, end_of_frame)



class TeleinfoException(Exception):
    pass


class ParserException(TeleinfoException):
    pass


class InvalidChecksumException(ParserException):
    pass


class UnknownKeyException(ParserException):
    def __init__(self, key):
        self.key = key
        self.message = "The key {key} is not known"



from serial_teleinfo.client import Client
from serial_teleinfo.exception import (
    TeleinfoException,
    ParserException,
    InvalidChecksumException,
    UnknownKeyException,
)
from serial_teleinfo.parser import parse_line
from serial_teleinfo.types import Value, type_parsers
from serial_teleinfo.updater import ValueUpdater

__all__ = [
    "Client",
    "TeleinfoException",
    "ParserException",
    "InvalidChecksumException",
    "UnknownKeyException",
    "parse_line",
    "Value",
    "type_parsers",
    "ValueUpdater",
]



import asyncio
from aiohttp import web
from typing import Dict

from serial_teleinfo.server.routes import routes
from serial_teleinfo import ValueUpdater


async def start_value_updater(app):
    app["updater"].start()


async def stop_value_updater(app):
    loop = asyncio.get_running_loop()
    await loop.run_in_executor(None, app["updater"].stop)


def app_factory(device: str, users: Dict[str, str]):
    app = web.Application()
    app.add_routes(routes)

    app["updater"] = ValueUpdater(device)
    app.on_startup.append(start_value_updater)
    app.on_cleanup.append(stop_value_updater)

    app["users"] = users

    return app



from serial_teleinfo.exception import ParserException
from typing import Callable, List


class Value:
    def __init__(
        self, key: str, value: any, unit: str = None, end_of_frame: bool = False
    ):
        self.key = key
        self.value = value
        self.unit = unit
        self.end_of_frame = end_of_frame

    def __str__(self):
        end_of_frame = " (END OF FRAME)" if self.end_of_frame else ""
        if self.unit is None:
            return f'<{self.key}="{self.value}"{end_of_frame}>'
        else:
            return f"<{self.key}={self.value} {self.unit}{end_of_frame}>"


def value_converter(
    lengths: List[int], parser: Callable[[str], any] = None, unit: str = None
) -> Callable[[str, str], Value]:
    if unit is not None:
        if parser is not None:
            raise ValueError("Cannot provide an unit and a parser")

        parser = int

    if parser is None:
        parser = str

    def _value_converter(key: str, value: str, end_of_frame: bool = False) -> Value:
        value_length = len(value)
        if value_length not in lengths:
            raise ParserException(
                f"Expected value length in {lengths}, got {value_length}"
            )

        return Value(key, parser(value), unit, end_of_frame)

    return _value_converter


def parse_str_remove_ellipsis(value: str) -> str:
    return value.rstrip(".")


# Extracted from the official documentation
# https://www.enedis.fr/sites/default/files/Enedis-NOI-CPT_02E.pdf
#
# dict containing (possible lengths, parser, unit) of each key
#
# Does not support :
# - "Compteur Jaune electronique (CJE)"
# - "Interface Clientele Emeraude a deux quadrants (ICE-2Q)"
# - "Interface Clientele Emeraude a quatre quadrants (ICE-4Q)"
# - "Compteur PME-PMI"
# - "Compteur SAPHIR"
type_parsers = {
    # Page 23
    "ADCO": value_converter([12]),
    "OPTARIF": value_converter([4], parse_str_remove_ellipsis),
    "BASE": value_converter([8, 9], unit="Wh"),
    "HCHC": value_converter([8, 9], unit="Wh"),
    "HCHP": value_converter([8, 9], unit="Wh"),
    "EJPHN": value_converter([8, 9], unit="Wh"),
    "EJPHPM": value_converter([8, 9], unit="Wh"),
    "GAZ": value_converter([7], unit="dal"),
    "AUTRE": value_converter([7], unit="dal"),
    "PTEC": value_converter([4], parse_str_remove_ellipsis),
    "MOTDETAT": value_converter([6]),
    # Page 26
    "ISOUSC": value_converter([2], unit="A"),
    "BBRHCJB": value_converter([9], unit="Wh"),
    "BBRHPJB": value_converter([9], unit="Wh"),
    "BBRHCJW": value_converter([9], unit="Wh"),
    "BBRHPJW": value_converter([9], unit="Wh"),
    "BBRHCJR": value_converter([9], unit="Wh"),
    "BBRHPJR": value_converter([9], unit="Wh"),
    "PEJP": value_converter([2], unit="min"),
    "DEMAIN": value_converter([4], parse_str_remove_ellipsis),
    "IINST": value_converter([3], unit="A"),
    "ADPS": value_converter([3], unit="A"),
    "IMAX": value_converter([3], unit="A"),
    "HHPHC": value_converter([1]),
    # page 30
    "PAPP": value_converter([5], unit="VA"),
    # page 32
    "IINST1": value_converter([3], unit="A"),
    "IINST2": value_converter([3], unit="A"),
    "IINST3": value_converter([3], unit="A"),
    "IMAX1": value_converter([3], unit="A"),
    "IMAX2": value_converter([3], unit="A"),
    "IMAX3": value_converter([3], unit="A"),
    "PMAX": value_converter([5], unit="W"),
    "PPOT": value_converter([2]),
}



from aiohttp import web
from http_basic_auth import parse_header, BasicAuthException

routes = web.RouteTableDef()


@routes.get("/status.json")
async def status(request):
    # Chech authorization
    try:
        authorization = request.headers.get("Authorization", "")
        login, password = parse_header(authorization)

        if login not in request.app["users"] or request.app["users"][login] != password:
            raise BasicAuthException()
    except BasicAuthException:
        raise web.HTTPUnauthorized(headers={"WWW-Authenticate": "Basic"})

    # Fetch and return data
    client = request.app["updater"]
    json_values = {
        key: [value.value, value.unit] for key, value in client.values.items()
    }

    return web.json_response({"connected": client.connected, "values": json_values})



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="serial-teleinfo",
    version="1.0.0",
    author="Ugo Méda",
    author_email="meda.ugo@gmail.com",
    description="Utilities to connect to a Teleinformation serial device",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ugomeda/serial-teleinfo",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
    extras_require={"server": ["aiohttp", "http_basic_auth"]},
    install_requires=["pyserial"],
)



import re
from typing import Tuple

from serial_teleinfo.exception import ParserException, InvalidChecksumException


SEPARATOR = b"\x20"
MATCHER = re.compile(b"^([^\x20\x09]+)([\x20\x09])(.+)([\x20\x09])([\x20-\x7e])$")


def parse_line(line: bytes) -> Tuple[str, str, bool]:
    """ Parses a line and verifies the checksum according to the specification
    available at https://www.enedis.fr/sites/default/files/Enedis-NOI-CPT_02E.pdf.

    Returns:
        (str, str, bool): A tuple containing the key, its value and
            a boolean at True if this is the end of a frame

    Raises:
        InvalidChecksumException: If the checksum was not correct
        ParserException: If separators could not be matched
    """
    # Remove markers
    end_of_frame = line.endswith(b"\r\x03\x02\n")
    if end_of_frame:
        line = line[0:-4]
    elif line.endswith(b"\r\n"):
        line = line[0:-2]
    else:
        raise ParserException(f"Could not find markers in {line}")

    # Split the line into its parts
    match = MATCHER.match(line)
    if match is None:
        raise ParserException(f"Could not extract data from {line}")

    key = match.group(1)
    separator_1 = match.group(2)
    value = match.group(3)
    checksum = match.group(5)

    # Verify the checksum
    checksum_sum = sum(key) + sum(separator_1) + sum(value)
    checksum_value = (checksum_sum & 0x3F) + 0x20

    if ord(checksum) != checksum_value:
        raise InvalidChecksumException(
            f"Expected checksum of {checksum_value}, found {checksum}"
        )

    return key.decode("ASCII"), value.decode("ASCII"), end_of_frame
